#!/bin/sh
#
# transplant_jbrowse_extraction.sh
#
# Script used to extract data from a JBrowse configuration.
#
# Author: E. Kimmel, R. Flores
#
# Copyright INRA-URGI 2016
#

test_return_code() {
  CODE=$1
  MESSAGE=$2
  if [ $CODE -ne 0 ];then
    echo "Error $CODE occured!"
    [ -n "${MESSAGE}" ] && echo "${MESSAGE}"
    echo "Exiting."
    exit 1
  fi
}

# Source: https://gist.github.com/cdown/1163649
urlencode() {
    # urlencode <string>
    old_lc_collate=$LC_COLLATE
    LC_COLLATE=C
    
    local length="${#1}"
    for (( i = 0; i < length; i++ )); do
        local c="${1:i:1}"
        case $c in
            [a-zA-Z0-9.~_-]) printf "$c" ;;
            *) printf '%%%02X' "'$c" ;;
        esac
    done
    
    LC_COLLATE=$old_lc_collate
}

# Source: https://gist.github.com/cdown/1163649
urldecode() {
    # urldecode <string>

    local url_encoded="${1//+/ }"
    printf '%b' "${url_encoded//%/\\x}"
}

check_jq() {
	which jq > /dev/null
	[ $? -ne 0 ] && echo "JQ is not installed on this server, or not specified in PATH: " && echo "$PATH" && echo "Please install it via package manager or via tarball: https://stedolan.github.io/jq/download/. Exiting." && exit 1

	JQ_VERSION=$(jq --version)
	[ "${JQ_VERSION}" != "jq-1.5" ] && echo "JQ version expected is jq-1.5. Please use this version only. Exiting." && exit 1
}
check_jq

MAX_CPU=$(grep -c processor /proc/cpuinfo)
test_return_code $?

# CURRENT_DIR=$(pwd)
CURRENT_DIR=$(dirname $0)

THREAD_NUMBER=${MAX_CPU}
WAIT=1

WORK_DIR="/tmp/index_jbrowse"
TAXON=""
SOURCE_NAME="GnpIS JBrowse"
APPLICATION_URL="https://urgi.versailles.inra.fr/jbrowse/gmod_jbrowse/?data=myData"

VERBOSE=0

## Help display
###############

help4me () {
    cat <<EOF

USAGE: 
  $0 -data_dir <data_dir> -conf_path <conf_path> -taxon <taxon> [-output_dir <output_dir>] [-source_name <index_source_name>] [-application_url <prefix_application_url>] [-thread_number <thread_number>] [--no_wait] [-v[v[v]]]

DESCRIPTION: 
  Script used to extract data from JBrowse indexed files.

PARAMS:
  -data_dir         the directory containing JBrowse files
  -conf_path        the JBrowse configuration path to extract data from
  -taxon            the value of the species field of the index
  -output_dir       the output directory where CSV (and temporary) files are generated (DEFAULT: ${WORK_DIR})
  -source_name      the name of the source field of the index (DEFAULT: ${SOURCE_NAME})
  -application_url  the prefix url of the application (DEFAULT: ${APPLICATION_URL})
  -thread_number    set the number of threads to use, enable parallelization of reference sequence extraction if the given number is greater than 1 (DEFAULT: ${MAX_CPU}, all available CPU)
  --no_wait         wait (1) or not (0) until all subjobs are finished, before exiting with success (DEFAULT: ${WAIT}). Useful only on multi thread mode (see above).
  -v                display verbose informations
  -vv               display very verbose informations
  -vvv              display very very verbose informations
  -h or --help      print this help

EOF
  exit 1
}

## Get commandline params
#########################

# get params
while [ -n "$1" ]; do
    case $1 in
        -h) help4me;shift 1;;
        --help) help4me;shift 1;;
        -data_dir) JBROWSE_DATA_DIR=$2;shift 2;;
        -conf_path) CONF_PATH=$2;shift 2;;
        -taxon) TAXON=$2;shift 2;;
        -output_dir) WORK_DIR=$2;shift 2;;
        -source_name) SOURCE_NAME=$2;shift 2;;
        -application_url) APPLICATION_URL=$2;shift 2;;
        -thread_number) THREAD_NUMBER=$2;shift 2;;
        --no_wait) WAIT=0;shift 1;;
        -v) VERBOSE=1;shift 1;;
        -vv) VERBOSE=2;shift 1;;
        -vvv) VERBOSE=3;shift 1;;
        --) shift;break;;
        -*) echo && echo "Unknown option: $1" && echo;exit 1;;
        *) break;;
    esac
done

[ -z "${JBROWSE_DATA_DIR}" ] && echo "ERROR: missing -data_dir option" && help4me && exit 1
[ -z "${CONF_PATH}" ] && echo "ERROR: missing -conf_path option" && help4me && exit 1
[ -z "${TAXON}" ] && echo "ERROR: missing -taxon option" && help4me && exit 1
[ ${THREAD_NUMBER} -lt 1 ] && echo "ERROR: -thread_number option must be greater or equal to 1" && help4me && exit 1
[ ${THREAD_NUMBER} -gt ${MAX_CPU} ] && echo "WARNING: -thread_number was set to ${THREAD_NUMBER} which is greater than available CPU, reducing it to ${MAX_CPU}." && THREAD_NUMBER=${MAX_CPU}
# [ -z "${WORK_DIR}" ] && echo "Missing -output_dir option" && help4me && exit 1

DATA_DIR="${JBROWSE_DATA_DIR}${CONF_PATH}"
[ ! -d ${WORK_DIR} ] && echo "Creating missing directory ${WORK_DIR}." && mkdir ${WORK_DIR}

flatten() {
	# Takes a UNIX path and flatten it, replacing all slash by underscore
	# Param 1 the path to flatten
	# Echoes the flattened path

	local flatten
	flatten=$(echo "$1" | sed -r 's/\//_/g')
	test_return_code $? "Cannot flatten path: '$1'."
	echo "$flatten"
	return 0
}

extractFeatures() {
		local ref_seq_dir="$1"
		local track="$2"
		local ref_seq url_encoded_track
		ref_seq=$(basename "$ref_seq_dir")
		url_encoded_track=$(urlencode "${track}")

		[ ${VERBOSE} -ge 2 ] && echo "Processing reference sequence ${ref_seq} for track ${track}..." # "$ref_seq_dir"

		local ref_seq_flattened
		ref_seq_flattened=$(flatten "${ref_seq_dir}")

		local classes_file
		classes_file="${WORK_DIR}/${ref_seq_flattened}__classes.json"
		jq '[.intervals.classes[]]' "${ref_seq_dir}"/trackData.json > "${classes_file}"
		test_return_code $?

		local has_chunk_files # checks weither the track stores features inside trackData.json or into LazyNCList format in chunk files
		# details: http://gmod.org/wiki/JBrowse_Configuration_Guide#Lazy_Nested-Containment_Lists_.28LazyNCList.29
		has_chunk_files=$(ls -1 "${ref_seq_dir}"/lf-*.json 2> /dev/null | wc -l)

		local json_source_file
		if [ "${has_chunk_files}" -eq 0 ]; then
		 	json_source_file=("${WORK_DIR}"/${ref_seq_flattened}__features.json)
			jq '[.intervals.nclist[]]' "${ref_seq_dir}"/trackData.json > "${json_source_file}"
		else
			json_source_file=("${ref_seq_dir}"/lf-*.json)
		fi
		[ ${VERBOSE} -ge 3 ] &&  echo "Source files are: " "${json_source_file[@]}"

		jq --slurpfile classes "${classes_file}" -f "${CURRENT_DIR}/maps_attributes_and_values.jq" "${json_source_file[@]}" \
		| sed -e ':a' -e 'N' -e '$!ba' -e 's/\n/, /g' \
		| sed -r 's/^(.*)$/[\1]/' \
		| jq -f "${CURRENT_DIR}/groups_features_on_null_separator.jq" \
		| jq -f "${CURRENT_DIR}/transforms_to_transplant_like_array.jq" --arg source_name "${SOURCE_NAME}" --arg track "${track}" --arg url_encoded_track "${url_encoded_track}" --arg taxon "${TAXON}" --arg application_url "${APPLICATION_URL}" --arg conf_path "${CONF_PATH}" \
		| jq -rs '.[]|@csv' \
		| sort | uniq \
		> ${WORK_DIR}/${ref_seq_flattened}__features.csv
		CODE=$?
		[ $CODE -gt 0 ] && echo "WARNING: problem $CODE occured for reference sequence $ref_seq"
		# 1st sed replaces new lines by commas https://stackoverflow.com/questions/1251999/how-can-i-replace-a-newline-n-using-sed
		# 2nd sed surrounds all lines by square brackets

		[ ${VERBOSE} -ge 2 ] && echo "Features for reference sequence ${ref_seq} of track ${track} extracted into ${WORK_DIR}/${ref_seq_flattened}__features.csv"
		rm -f "${WORK_DIR}/${ref_seq_flattened}__"*.json # remove all temporary json files
}

JOBS_NB=0
for track_dir in ${DATA_DIR}/tracks/* ;
do 
	track=$(basename "$track_dir")
	track_flattened=$(flatten "${track_dir}")
	[ ! -d "${track_dir}" ] && 
		{ # in case of file system error such as mount point deleted or else
			echo "ERROR: directory ${track} became unavailable. Exiting." && exit 2
		}
	[ ${VERBOSE} -ge 1 ] && echo "Processing track: ${track}... " # "$track_dir"

	for ref_seq_dir in "${track_dir}"/*
	do
		
		# if [ $(echo "$ref_seq_dir" | grep -c "ARRAY") -eq 0 ] && [ $(echo "$ref_seq_dir" | grep -c "Marker") -eq 0 ] ; then
		# 	break # testing purpose only, delete this 3 lines block when array mapping will be OK in maps_attributes_and_values.jq
		# fi

		if [ ${THREAD_NUMBER} -gt 1 ]; then
			while [ ${JOBS_NB} -ge ${THREAD_NUMBER} ];
			do 
				JOBS_NB=$(jobs | wc -l)
				[ ${VERBOSE} -ge 4 ] && echo "Waiting previous ${JOBS_NB} processes finish..."
				[ ${VERBOSE} -ge 4 ] && jobs
				sleep 0.01
			done
			[ ! -d "${ref_seq_dir}" ] && 
				{ # in case of file system error such as mount point deleted or else
					echo "ERROR: directory ${ref_seq_dir} became unavailable. Exiting." && exit 2
				}
			(extractFeatures "${ref_seq_dir}" "${track}")&
			JOBS_NB=$(jobs | wc -l)
		else
			extractFeatures "${ref_seq_dir}" "${track}"
		fi
	done
	pid=($(jobs -p))
	if [ ${THREAD_NUMBER} -gt 1 ] && [ ${#pid[@]} -gt 0 ]; then
			while [ $(ps hp "${pid[@]}" | wc -l) -gt 0 ]; # should care if one of $pid is reassigned to a new process by kernel
			do
				sleep 0.1 # wait 0.1 second to check if processes are terminated or not.
			done
	fi
	(
		cat "${WORK_DIR}/${track_flattened}"*.csv | gzip -v -1 -c > "${WORK_DIR}/${track_flattened}_features.csv.gz" 2> "${WORK_DIR}/${track_flattened}.percent"
		[ ${VERBOSE} -ge 1 ] && echo "Combined and compressed by$(cat "${WORK_DIR}/${track_flattened}".percent) all reference sequences for track (${track}) into: ${WORK_DIR}/${track_flattened}_features.csv.gz"
		rm -f "${WORK_DIR}/${track_flattened}"*.csv "${WORK_DIR}/${track_flattened}.percent"
	) &
done

# Wait until all subjobs are finished, before exiting with success.
if [ ${WAIT} -eq 1 ] && [ ${THREAD_NUMBER} -gt 1 ]; then
	LAST_JOBS_NB=$(jobs | wc -l)
	[ ${VERBOSE} -ge 1 ] && echo "Waiting last ${LAST_JOBS_NB} processes finish..."
	wait
fi

[ ${VERBOSE} -ge 1 ] && echo "Extraction of $CONF_PATH finished."

exit 0
