#!/bin/bash
#
# index_solr.sh
#
# Admin script used to extract data from a GnpIS thematic.
#
# Author: E. Kimmel, R. Flores
#
# Copyright INRA-URGI 2014
#

# default values
CORE_NAME="URGI"
DIR=
USER=
PASSWORD=
SERVER_NAME=dev
HOST_DEV="gimli.versailles.inra.fr"
PORT="8993"
DO_CLEAN=
DO_OPTIMIZE=
DO_INDEX=1
VERBOSE=0
SERVER_PROD="https://urgi.versailles.inra.fr"
STREAM=0

help4index_solr() {
    cat <<EOF
Usage:
	$0 [-d <DIR>] [-u <USER>] [-p <PASSWORD>] -s [<dev>|<prod>] [-c <CORE_NAME>] [-host <localhost>] [-port <8993>] [--clean] [--optimize] [--index <0|1>] [--stream] [-v[v]]

DESCRIPTION: 
	Script used to launch GnpIS data indexing in SolR.

PARAMS:
	-c			the core used on SolR (DEFAULT: ${CORE_NAME})
	-d			the input directory which contains data file (tabulated files) to index
	-host		the host used for solr dev instance (DEFAULT: ${HOST_DEV})	
	-u			the user to use to log in SolR admin interface
	-p			the password to use to log in SolR admin interface
	-port		the port used by the solr instance (DEFAULT: ${PORT})
	-s			the server to index on ([dev]|prod)
	--clean		clean first all indices on SolR
	--optimize	optimize index on SolR
	--index		according to option given (1 OR 0), index (1) or not (0) data (allowing to clean or optimize only without indexing). Default: ${DO_INDEX}.
	--stream	index data in a streaming fashion using directly STDIN as data source to index on SolR.
	-v			verbose mode
	-vv			very verbose mode
EOF
  exit 1
}

# any params
[ -z "$1" ] && echo && help4index_solr

# get params
while [ -n "$1" ]; do
    case $1 in
        -h) help4index_solr;shift 1;;
        --help) help4index_solr;shift 1;;
		-c) CORE_NAME=$2;shift 2;;
        -d) DIR=$2;shift 2;;
		-host) HOST_DEV=$2;shift 2;;
        -u) USER=$2;shift 2;;
        -p) PASSWORD=$2;shift 2;;
		-port) PORT=$2;shift 2;;
        -s) SERVER_NAME=$2;shift 2;;
        --clean) DO_CLEAN=$1;shift 1;;
        --optimize) DO_OPTIMIZE=$1;shift 1;;
        --index) DO_INDEX=$2;shift 1;;
        --stream) STREAM=1;shift 1;;
        -v) VERBOSE=1;shift 1;;
        -vv) VERBOSE=2;shift 1;;
        --) shift;break;;
        -*) echo && echo "Unknown option: $1" && echo;exit 1;;
        *) break;;
    esac
done

SERVER_DEV="http://${HOST_DEV}:${PORT}"

check_error() {
  CODE=$1
  if [ "$CODE" -ne 0 ];then
    echo "Error $CODE occured!"
    echo "Exiting."
    exit 1
  fi
}

echo_and_eval_cmd() {
  local CMD=$1
  if [ -z "$CMD" ]; then
    echo "Missing command to eval. Exiting."
    exit 1
  fi
  PRINTED_CMD="${CMD}"
  if [ -n "${CREDENTIALS}" ]; then
  	PRINTED_CMD=$(echo "${CMD}" | sed -re "s/${CREDENTIALS}/-u ***:***/g")
  fi
  [ $VERBOSE -ge 1 ] && echo -e "Going to exec command: \n\t${PRINTED_CMD}"
  eval "$CMD"
  check_error $?
}

if [ -z "$DIR" ] && [ $STREAM -eq 0 ]; then
	DIR="."
	echo "Assuming that data files are in current directory..." && sleep 3 # waiting 3s to allow stop process if directory parameters was missing
fi

if [[ ( -z "$USER" ) && ( "$SERVER_NAME" == "prod" ) ]]; then
	echo "Missing production SolR admin web username, exiting..."
	help4index_solr
fi

if [ -z "$SERVER_NAME" ]; then
	echo "Server name has not been specified, using dev server by default."
fi

SERVER=${SERVER_DEV}
CREDENTIALS=
if [ -n "${SERVER_NAME}" ]; then
	if [ "${SERVER_NAME}" == 'prod' ]; then
		SERVER=${SERVER_PROD}
		if [ -z "$PASSWORD" ]; then
			read -s -p "SolR admin web password? " PASSWORD
		fi
		echo
		if [ -z "$USER" ]; then
			echo "SolR admin web password could not be empty, exiting..."
			help
		fi
		CREDENTIALS="-u $USER:$PASSWORD"
	fi
fi
echo "Using server: " ${SERVER}

SOLR="solr"
# SOLR_CORE="transplant-0.2-pre"
SOLR_CORE="${CORE_NAME}"
SOLR_COMMAND_DELETE="update?commit=true"
SOLR_COMMAND_OPTIMIZE="update?optimize=true&waitFlush=true"
SOLR_COMMAND_INDEX="update/csv"
#SEPARATOR="%09"
SEPARATOR="%2C"
FIELDNAMES="entry_type,\
database_name,\
db_id,\
db_version,\
description,\
url,\
species,\
xref,\
feature_type,\
sequence_id,\
sequence_version,\
start_position,\
end_position,\
map,\
map_position,\
authority,\
trait,\
trait_id,\
environment,\
environment_id,\
statistic,\
unit,\
genotype,\
experiment_type"
HEADER="false"
OVERWRITE="true"
CONTENT_TYPE_XML="Content-type:text/xml; charset=utf-8"
CONTENT_TYPE_CSV="Content-type:application/csv; charset=utf-8"
CURL="curl -i"
CONNECT_COMMAND="$CURL $CREDENTIALS $SERVER/$SOLR"
DELETE_COMMAND="$CURL $CREDENTIALS $SERVER/$SOLR/$SOLR_CORE/$SOLR_COMMAND_DELETE -d '<delete><query>*:*</query></delete>' -H '$CONTENT_TYPE_XML'"
OPTIMIZE_COMMAND="$CURL $CREDENTIALS $SERVER/$SOLR/$SOLR_CORE/$SOLR_COMMAND_OPTIMIZE"
COMMIT_COMMAND=""
INDEX_COMMAND="$CURL $CREDENTIALS \
'$SERVER/$SOLR/$SOLR_CORE/$SOLR_COMMAND_INDEX?separator=$SEPARATOR&fieldnames=$FIELDNAMES&header=$HEADER&overwrite=$OVERWRITE&commit=true&f.species.split=true&f.species.separator=$SEPARATOR' \
-H '$CONTENT_TYPE_CSV' --data-binary @"

# clean previous indices
clean() {
	echo "Clean previous indices..."
	echo_and_eval_cmd "$DELETE_COMMAND"
	echo
}

optimize() {
	echo "Optimize index..."
	echo_and_eval_cmd "$OPTIMIZE_COMMAND"
	echo
}

# index *.csv files
index() {
	echo "Index *.csv files in $DIR..."
	cd $DIR || exit 1
	for csv_file in *.csv; do
		[ ! -f "${csv_file}" ] && echo "${csv_file} does not exist. Exiting." && exit 1
		local PREFIX="part_${csv_file}"
		#local SUFFIX="_${csv_file}"
		split -l 1000000 "${csv_file}" "${PREFIX}" # --additional-suffix="${SUFFIX}"
		for split in ./"${PREFIX}"*"${SUFFIX}"; do
			echo " * Indexing ${split}..."
			echo_and_eval_cmd "cat '${split}' | grep -v '^#' | ${INDEX_COMMAND}/dev/stdin | grep -v '<int name="status">400</int>'"
			echo
			rm "${split}"
		done
	done
	cd - || exit 1
}

index_stream() {
	echo " * Indexing via stdin..."
	echo_and_eval_cmd "cat | grep -v '^#' | ${INDEX_COMMAND}/dev/stdin | grep -v '<int name="status">400</int>'"
}

testConnection() {
	if [ "${SERVER_NAME}" == 'prod' ]; then
		echo_and_eval_cmd "${CONNECT_COMMAND} | grep 302"
	fi
}

testConnection # check if SolR server is running and if we can connect to it

if [ "${DO_CLEAN}" == "--clean" ]; then
	clean
fi

if [ "${DO_INDEX}" -eq 1 ]; then
	if [ "${STREAM}" == "1" ]; then
		index_stream
	else
		index
	fi
fi

if [ "${DO_OPTIMIZE}" == "--optimize" ]; then
	optimize
fi

exit 0
